import random
import pyglet
pyglet.options['debug_gl'] = False
#def zero_errcheck( result, func, arguments ):		# die, error checking, die!
#	return result
#pyglet.gl.lib.errcheck = zero_errcheck

from pyglet.gl import *
pyglet.resource.path = ['data']
pyglet.resource.reindex()
pyglet.resource.add_font( 'Tuffy.ttf' )
pyglet.resource.add_font( 'Tuffy_Bold.ttf' )

# psyco seems to kill the pyglet mainloop
#try:
#	import psyco
#	psyco.full()
#except:
#	pass

GRIDX, GRIDY = 190, 160
STEPTIME = 20
fSTEPTIME = float( STEPTIME )

class Game( object ):
	def __init__( self ):
		self.option_childvision = True
		self.option_tooltips = True
		self.option_edgescroll = False

		self.music = None
		self.setmusic( 'guitar.ogg' )
		self.phase = Menu( self )

		self.window = pyglet.window.Window( width=1024, height=640, caption='LAWN: Terror From The Green' )
		self.window.set_icon( pyglet.resource.image( 'icon16.png' ).get_image_data(), pyglet.resource.image( 'icon32.png' ).get_image_data(), pyglet.resource.image( 'icon128.png' ).get_image_data() )
		@self.window.event
		def on_draw():
			self.window.clear()
			self.phase.draw()
			self.window.invalid = False
		@self.window.event
		def on_key_press( symbol, modifiers ):
			self.phase.on_key_press( symbol, modifiers )
		@self.window.event
		def on_key_release( symbol, modifiers ):
			self.phase.on_key_release( symbol, modifiers )
		@self.window.event
		def on_mouse_motion( x, y, dx, dy ):
			self.phase.on_mouse_motion( x, y, dx, dy )
		@self.window.event
		def on_mouse_press( x, y, button, modifiers ):
			self.phase.on_mouse_press( x, y, button, modifiers )
		@self.window.event
		def on_mouse_release( x, y, button, modifiers ):
			self.phase.on_mouse_release( x, y, button, modifiers )
		@self.window.event
		def on_mouse_drag( x, y, dx, dy, button, modifiers ):
			self.phase.on_mouse_drag( x, y, dx, dy, button, modifiers )
		@self.window.event
		def on_mouse_scroll( x, y, scroll_x, scroll_y ):
			self.phase.on_mouse_scroll( x, y, scroll_x, scroll_y )
		def think( dt ):
			self.phase.think()
			self.window.invalid = True
		self.phase.init()
		pyglet.clock.schedule_interval( think, 1 / 40.0 )
	def setmusic( self, music ):
		if self.music is not None:
			self.music.pause()
		self.music = pyglet.media.Player()
		try:
			self.music.queue( pyglet.resource.media( music ) )
			self.music.eos_action = self.music.EOS_LOOP
		except:
			print 'AVbin is missing or does not support Ogg/Vorbis, so music is out. Hum something!'
	def start( self ):
		pyglet.app.run()
	def setphase( self, phase ):
		self.phase.end()
		self.phase = phase
		phase.init()

class Phase( object ):
	def draw( self ):
		pass
	def think( self ):
		pass
	def init( self ):
		pass
	def end( self ):
		pass
	def on_key_press( self, symbol, modifiers ):
		pass
	def on_key_release( self, symbol, modifiers ):
		pass
	def on_mouse_motion( self, x, y, dx, dy ):
		pass
	def on_mouse_press( self, x, y, button, modifiers ):
		pass
	def on_mouse_release( self, x, y, button, modifiers ):
		pass
	def on_mouse_drag( self, x, y, dx, dy, button, modifiers ):
		pass
	def on_mouse_scroll( self, x, y, scroll_x, scroll_y ):
		pass

class World( object ):
	@staticmethod
	def random():
		w = World()
		w.deployment = None
		return w

class Transition( Phase ):
	def __init__( self, game, a, b ):
		self.game = game
		self.a = a
		self.b = b
	def think( self ):
		self.game.setphase( self.b )

class Base( Phase ):
	def __init__( self, game ):
		self.game = game
		self.world = game.world
	def think( self ):
		self.world.deployment = TestDeployment()
		self.game.setphase( Transition( self.game, self, Action( self.game ) ) )

class TestDeployment( object ):
	def __init__( self ):
		self.width = 10
		self.height = 10

imgcache = {}
def sprite( imgname, batch = None, mipmap = True ):
	# probably pyglet also does caching but who knows
	if imgname not in imgcache:
		image = pyglet.resource.image( imgname )
		if mipmap:
			image = image.get_image_data()
			image.get_regular_texture = image.get_texture
			image.get_texture = image.get_mipmapped_texture
		imgcache[imgname] = image
	try:
		return pyglet.sprite.Sprite( imgcache[imgname], batch = batch )
	except pyglet.image.ImageException:
		imgcache[imgname].get_texture = imgcache[imgname].get_regular_texture
		return pyglet.sprite.Sprite( imgcache[imgname], batch = batch )

sndcache = {}
def playsound( sndname ):
	return pyglet.resource.media( sndname ).play()

class Actor( object ):
	def __init__( self, world ):
		self.world = world
		self.setpos( 1, 1 )
		self.setfacing( 'down' )
		self.animations = []
		self.sight = 0
		self.opaque = False
		self.sound = None
		self.visible = True
		self.opacity = 255
	def playsound( self, sndname ):
		if self.sound is None or self.sound.time == 0:		# if the sound finished playing time is set to 0 (nicer way to tell?)
			self.sound = playsound( sndname )
	def setpos( self, x, y ):
		self.x = self.visiblex = x
		self.y = self.visibley = y
	def setfacing( self, facing ):
		self.facing = self.visiblefacing = facing
	def think( self ):
		for anim in self.getanimation():
			anim.think()
		if self.visible and self.opacity < 255:
			self.opacity += 5
		elif not self.visible and self.opacity > 0:
			self.opacity -= 5
	def drawshadowsprite( self ):
		sprite = getattr( self, self.visiblefacing + '_shadow' )
		sprite.draw()
	def drawshadow( self ):
		sprite = getattr( self, self.visiblefacing + '_shadow' )
		glPushMatrix()
		glTranslatef( self.visiblex * GRIDX - sprite.width/2 + GRIDX/2, self.visibley * GRIDY - sprite.height/2 + GRIDY/2, 0 )
		for anim in self.getanimation():
			anim.animate_shadow( sprite )
		sprite.opacity = sprite.opacity * self.opacity / 255
		self.drawshadowsprite()
		sprite.opacity = 255
		glPopMatrix()
	def draw( self ):
		sprite = getattr( self, self.visiblefacing )
		glPushMatrix()
		glTranslatef( self.visiblex * GRIDX - sprite.width/2 + GRIDX/2, self.visibley * GRIDY - sprite.height/2 + GRIDY/2, 0 )
		for anim in self.getanimation():
			anim.animate( sprite )
		sprite.opacity = sprite.opacity * self.opacity / 255
		self.drawsprite()
		sprite.opacity = 255
		glPopMatrix()
	def drawsprite( self ):
		sprite = getattr( self, self.visiblefacing )
		sprite.draw()
	def vision( self ):
		dx, dy = facing_to_offset( self.facing )
		opaques = []
		cone = []
		ax, ay = self.x, self.y
		map = self.world.map
		for i in range( 1, self.sight + 1 ):
			for j in range( i * 2 + 1 ):
				x = ax + i * dx + (i - j) * dy
				y = ay + i * dy + (i - j) * dx
				cone.append( (x, y) )
				if (x, y) in map and map[x, y].opaque:
					opaques.append( (x, y) )
		for x, y in cone:
			if (x, y) == (ax, ay):
				yield x, y
				continue
			for ox, oy in opaques:
				da = ((x - ax) ** 2 + (y - ay) ** 2) ** 0.5
				dp = abs( (x - ax) * (ax - ox) + (ay - oy) * (y - ay) ) / da
				if dp > da:
					continue
				d = abs( (x - ax) * (ay - oy) - (ax - ox) * (y - ay) ) / da
				if d < 1:
					break
			else:
				yield x, y
	def addanimation( self, animname ):
		self.animations.append( getattr( self, animname )( self ) )
	def getanimation( self ):
		if self.animations:
			yield self.animations[0]
	def endanimation( self ):
		self.animations.pop( 0 )
	class SimpleChange( object ):
		def __init__( self, object ):
			self.object = object
			self.starttime = object.world.time
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.change()
			self.object.endanimation()
		def animate_shadow( self, sprite ):
			pass
		def animate( self, sprite ):
			pass
	class TurnLeft( SimpleChange ):
		def change( self ):
			self.object.visiblefacing = 'left'
	class TurnRight( SimpleChange ):
		def change( self ):
			self.object.visiblefacing = 'right'
	class TurnUp( SimpleChange ):
		def change( self ):
			self.object.visiblefacing = 'up'
	class TurnDown( SimpleChange ):
		def change( self ):
			self.object.visiblefacing = 'down'
	def step( self ):
		pass

def jumpy( x ):
	# like sine, but parabolic...
	import math
	x = (x % math.pi) / math.pi
	return 1 - 4 * (x-0.5) ** 2

class Child( Actor ):
	def __init__( self, world, properties ):
		Actor.__init__( self, world )
		self.properties = properties
		self.selected = False
		self.selected_animation = self.Selected( self )
		self.panicked = False
		self.panicked_animation = self.Panicked( self )
		self.panic = sprite( 'panic.png' )
		self.select_mark = sprite( 'selected.png' )
	def getanimation( self ):
		if self.animations:
			yield self.animations[0]
		elif self.panicked:
			yield self.panicked_animation
		elif self.selected:
			yield self.selected_animation
	def endanimation( self ):
		self.animations.pop( 0 )
		self.selected_animation.phase = 0
		self.panicked_animation.phase = 0
	def drawsprite( self ):
		sprite = getattr( self, self.visiblefacing )
		sprite.draw()
		glTranslatef( 0, GRIDY, 0 )
		if self.selected:
			self.select_mark.draw()
		if self.panicked:
			self.panic.draw()
	class Selected( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
		def animate_shadow( self, sprite ):
			import math
			h = 10 * abs( math.sin( self.phase ) )
			s = 1 / (0.02 * h + 1)
#			glTranslatef( sprite.width/2, sprite.height/2, 0 )
#			glScalef( s, s, 0 )
#			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
#			glTranslatef( -h, -h/2, 0 )
		def animate( self, sprite ):
			import math
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glRotatef( 1.5 * math.sin( self.phase ), 0, 0, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
#			glTranslatef( 0, 10 * abs( math.sin( self.phase ) ), 0 )
		def think( self ):
			self.phase += 0.2
	class Panicked( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
		def animate_shadow( self, sprite ):
			import math
			h = 10 * abs( jumpy( self.phase ) )
			s = 1 / (0.02 * h + 1)
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glScalef( s, s, 0 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			glTranslatef( -h, -h/2, 0 )
		def animate( self, sprite ):
			import math
			glTranslatef( 0, 10 * abs( jumpy( self.phase ) ), 0 )
		def think( self ):
			self.phase += 0.3
	class Move( object ):
		def __init__( self, object, dx, dy ):
			self.object = object
			self.phase = 0
			self.dx, self.dy = dx, dy
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			import math
			h = 30 * abs( jumpy( self.phase * math.pi / STEPTIME ) )
#			s = 1 / (0.02 * h + 1)
#			glTranslatef( sprite.width/2, sprite.height/2, 0 )
#			glScalef( s, s, 0 )
#			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			s = 1 / (0.1 * h + 1)
			sprite.opacity = s * 255
			glTranslatef( -h, -h/2, 0 )
			glTranslatef( self.dx * self.phase * GRIDX / fSTEPTIME, self.dy * self.phase * GRIDY / fSTEPTIME, 0 )
		def animate( self, sprite ):
			import math
			glTranslatef( self.dx * self.phase * GRIDX / fSTEPTIME, self.dy * self.phase * GRIDY / fSTEPTIME + 30 * abs( jumpy( self.phase * math.pi / STEPTIME ) ), 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase == 1:
				if isinstance( self.object, Girl ):
					playsound( 'stepsound-girl.wav' )
				else:
					playsound( 'stepsound.wav' )
			if self.phase >= STEPTIME:
				self.object.visiblex += self.dx
				self.object.visibley += self.dy
				self.object.endanimation()
	class MoveLeft( Move ):
		def __init__( self, object ):
			Boy.Move.__init__( self, object, -1, 0 )
	class MoveRight( Move ):
		def __init__( self, object ):
			Boy.Move.__init__( self, object, 1, 0 )
	class MoveUp( Move ):
		def __init__( self, object ):
			Boy.Move.__init__( self, object, 0, 1 )
	class MoveDown( Move ):
		def __init__( self, object ):
			Boy.Move.__init__( self, object, 0, -1 )
	class Use( object ):
		def __init__( self, object, dx, dy ):
			self.object = object
			self.phase = 0
			self.dx, self.dy = dx, dy
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			import math
			h = 30 * abs( jumpy( 2 * self.phase * math.pi / STEPTIME ) )
			s = 1 / (0.1 * h + 1)
			sprite.opacity = s * 255
			glTranslatef( -h, -h/2, 0 )
			p = STEPTIME / 2 - abs( self.phase - STEPTIME / 2 )
			glTranslatef( self.dx * p * GRIDX / fSTEPTIME, self.dy * p * GRIDY / fSTEPTIME, 0 )
		def animate( self, sprite ):
			import math
			p = STEPTIME / 2 - abs( self.phase - STEPTIME / 2 )
			glTranslatef( self.dx * p * GRIDX / fSTEPTIME, self.dy * p * GRIDY / fSTEPTIME + 30 * abs( jumpy( 2 * self.phase * math.pi / STEPTIME ) ), 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase >= STEPTIME:
				self.object.endanimation()
	class UseLeft( Use ):
		def __init__( self, object ):
			Boy.Use.__init__( self, object, -1, 0 )
	class UseRight( Use ):
		def __init__( self, object ):
			Boy.Use.__init__( self, object, 1, 0 )
	class UseUp( Use ):
		def __init__( self, object ):
			Boy.Use.__init__( self, object, 0, 1 )
	class UseDown( Use ):
		def __init__( self, object ):
			Boy.Use.__init__( self, object, 0, -1 )
	class Yikes( Actor.SimpleChange ):
		def change( self ):
			self.object.playsound( 'yikes-%d.wav'%random.randint( 1, 6 ) )
	class Escaped( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			sprite = getattr( self.object, self.object.visiblefacing + '_shadow' )
			sprite.opacity = 255 * (STEPTIME - self.phase) / STEPTIME
		def animate( self, sprite ):
			sprite = getattr( self.object, self.object.visiblefacing )
			sprite.opacity = 255 * (STEPTIME - self.phase) / STEPTIME
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase >= STEPTIME:
				self.object.endanimation()
				self.object.world.noncolliders.remove( self.object )
				self.object.world.addmessage( '%s has safely escaped!'%self.object.properties.name )
	def step( self ):
		if not self.panicked:
			return
		if self.justpanicked:		# start running around on next turn
			self.justpanicked = False
			return
		self.addanimation( 'Yikes' )
		dx, dy = facing_to_offset( self.facing )
		if not self.world.free( self.x + dx, self.y + dy ):
			for f in perpendicular_facing( self.facing ):
				dx, dy = facing_to_offset( self.facing )
				if self.world.free( self.x + dx, self.y + dy ):
					self.facing = f
					self.addanimation( 'Turn' + self.facing.capitalize() )
			else:
				self.facing = facing_cw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
		else:
			self.x += dx
			self.y += dy
			self.addanimation( 'Move' + self.facing.capitalize() )
			if not self.world.is_lawn( self.x, self.y ):
				self.world.objects.remove( self )
				self.world.children.remove( self )
				self.world.noncolliders.append( self )
				self.addanimation( 'Escaped' )
		self.world.noise( self.x, self.y, 10 )
	def can_act( self ):
		return not self.panicked

class Boy( Child ):
	def __init__( self, world, properties ):
		Child.__init__( self, world, properties )
		self.sight = properties.sight
		k = properties.kind
		self.left = sprite( 'child-%d-left.png'%k )
		self.right = sprite( 'child-%d-right.png'%k )
		self.down = sprite( 'child-%d-front.png'%k )
		self.up = sprite( 'child-%d-back.png'%k )
		self.left_shadow = sprite( 'child-%d-left-shadow.png'%k )
		self.right_shadow = sprite( 'child-%d-right-shadow.png'%k )
		self.down_shadow = sprite( 'child-%d-front-shadow.png'%k )
		self.up_shadow = sprite( 'child-%d-back-shadow.png'%k )
		self.portrait = sprite( 'child-%d-portrait.png'%k )

class Girl( Child ):
	def __init__( self, world, properties ):
		Child.__init__( self, world, properties )
		self.sight = properties.sight
		k = properties.kind
		self.left = sprite( 'girl-%d-left.png'%k )
		self.right = sprite( 'girl-%d-right.png'%k )
		self.down = sprite( 'girl-%d-front.png'%k )
		self.up = sprite( 'girl-%d-back.png'%k )
		self.left_shadow = sprite( 'girl-%d-left-shadow.png'%k )
		self.right_shadow = sprite( 'girl-%d-right-shadow.png'%k )
		self.down_shadow = sprite( 'girl-%d-front-shadow.png'%k )
		self.up_shadow = sprite( 'girl-%d-back-shadow.png'%k )
		self.portrait = sprite( 'girl-%d-portrait.png'%k )
		self.talkingto = None
	class SayHi( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.starttime = object.world.time
			self.balloon = sprite( 'balloon-girl-talk.png' )
			self.rainbow = sprite( 'rainbow.png' )
		def animate_shadow( self, sprite ):
			pass
		def animate( self, sprite ):
			if self.phase == 0:
				return
			import math
			s = math.sin( self.phase * math.pi / fSTEPTIME ) ** 0.2

			glPushMatrix()
			glTranslatef( (self.object.talkingto.visiblex - self.object.visiblex) * GRIDX, (self.object.talkingto.visibley - self.object.visibley) * GRIDY, 0 )
			glTranslatef( 0, 250, 0 )
			self.rainbow.opacity = 255 * s
			self.rainbow.draw()
			glPopMatrix()

			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glRotatef( 3 * math.sin( math.pi * self.phase * 2 / fSTEPTIME ), 0, 0, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			glPushMatrix()
			glTranslatef( 0, GRIDY, 0 )
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glScalef( s, s, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			self.balloon.draw()
			glPopMatrix()
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase == 1:
				self.object.playsound( 'girl-%d.wav'%random.randint( 1, 6 ) )
			if self.phase >= STEPTIME:
				self.object.endanimation()
	def talkto( self, elderly ):
		self.talkingto = elderly
	def step( self ):
		if self.talkingto is not None:
			self.talkingto.hear( self.x, self.y, 100 )	# distract him
			self.addanimation( 'SayHi' )
	def can_act( self ):
		return not self.panicked and self.talkingto is None

class Elderly( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.sight = 4
		self.left = sprite( 'elderly-left.png' )
		self.right = sprite( 'elderly-right.png' )
		self.down = sprite( 'elderly-front.png' )
		self.up = sprite( 'elderly-back.png' )
		self.left_shadow = sprite( 'elderly-left-shadow.png' )
		self.right_shadow = sprite( 'elderly-right-shadow.png' )
		self.down_shadow = sprite( 'elderly-front-shadow.png' )
		self.up_shadow = sprite( 'elderly-back-shadow.png' )
		self.timedivider = 1
		self.activity = 2
		self.interest = None
		self.justheard = False
	class Move( object ):
		def __init__( self, object, dx, dy ):
			self.object = object
			self.phase = 0
			self.dx, self.dy = dx, dy
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			glTranslatef( self.dx * self.phase * GRIDX * 0.5 / fSTEPTIME, self.dy * self.phase * GRIDY * 0.5 / fSTEPTIME, 0 )
		def animate( self, sprite ):
			import math
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glRotatef( 0.9 * math.sin( 4 * math.pi * self.phase / fSTEPTIME ), 0, 0, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			glTranslatef( self.dx * self.phase * GRIDX * 0.5 / fSTEPTIME, self.dy * self.phase * GRIDY * 0.5 / fSTEPTIME, 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase >= STEPTIME * 2:
				self.object.visiblex += self.dx
				self.object.visibley += self.dy
				self.object.endanimation()
	class MoveLeft( Move ):
		def __init__( self, object ):
			Elderly.Move.__init__( self, object, -1, 0 )
	class MoveRight( Move ):
		def __init__( self, object ):
			Elderly.Move.__init__( self, object, 1, 0 )
	class MoveUp( Move ):
		def __init__( self, object ):
			Elderly.Move.__init__( self, object, 0, 1 )
	class MoveDown( Move ):
		def __init__( self, object ):
			Elderly.Move.__init__( self, object, 0, -1 )
	class SayHi( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.starttime = object.world.time
			self.balloon = sprite( 'balloon-hi-there.png' )
		def animate_shadow( self, sprite ):
			pass
		def animate( self, sprite ):
			if self.phase == 0:
				return
			import math
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glRotatef( 3 * math.sin( math.pi * self.phase / fSTEPTIME ), 0, 0, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			glPushMatrix()
			glTranslatef( 0, GRIDY, 0 )
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			s = math.sin( self.phase * math.pi * 0.5 / fSTEPTIME ) ** 0.2
			glScalef( s, s, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			self.balloon.draw()
			glPopMatrix()
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase == 1:
				self.object.playsound( 'hithere.wav' )
			if self.phase >= STEPTIME * 2:
				self.object.endanimation()
	def hear( self, x, y, strength ):
		if self.interest is None or self.interest[2] < strength:
			self.interest = x, y, strength
			self.justheard = True
	def step( self ):
		hi = False
		for x, y in self.vision():
			o = self.world.map.get( (x, y), None )
			if o and isinstance( o, Child ) and not o.panicked:
				if isinstance( o, Girl ) and (self.x, self.y) in list( o.vision() ):
					o.talkto( self )
				else:
					o.panicked = True
					o.justpanicked = True
					hi = True
		if hi:
			self.addanimation( 'SayHi' )
			return

		self.timedivider -= 1
		if self.timedivider != 0: return
		self.timedivider = 2

		justheard = self.justheard
		self.justheard = False
		if self.interest is not None:
			ix, iy, s = self.interest
			if abs( ix - self.x ) > abs( iy - self.y ):
				if ix < self.x:
					if self.facing == 'left':
						self.interest = None
						if justheard: return
					elif self.facing == 'right':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'left'
						self.addanimation( 'TurnLeft' )
						return
				else:
					if self.facing == 'right':
						self.interest = None
						if justheard: return
					elif self.facing == 'left':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'right'
						self.addanimation( 'TurnRight' )
						return
			else:
				if iy < self.y:
					if self.facing == 'down':
						self.interest = None
						if justheard: return
					elif self.facing == 'up':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'down'
						self.addanimation( 'TurnDown' )
						return
				else:
					if self.facing == 'up':
						self.interest = None
						if justheard: return
					elif self.facing == 'down':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'up'
						self.addanimation( 'TurnUp' )
						return

		if self.activity > 0:
			r = random.randint( 1, 100 * (6 - self.activity) )
			if r <= 50:
				dx, dy = facing_to_offset( self.facing )
				if self.world.free_and_lawn( self.x + dx, self.y + dy ):
					self.x += dx
					self.y += dy
					self.addanimation( 'Move' + self.facing.capitalize() )
			elif r <= 75:
				self.facing = facing_cw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
			elif r <= 100:
				self.facing = facing_ccw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )

class Dog( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.left = sprite( 'dog-left.png' )
		self.right = sprite( 'dog-right.png' )
		self.down = sprite( 'dog-front.png' )
		self.up = sprite( 'dog-back.png' )
		self.left_shadow = sprite( 'dog-left-shadow.png' )
		self.right_shadow = sprite( 'dog-right-shadow.png' )
		self.down_shadow = sprite( 'dog-front-shadow.png' )
		self.up_shadow = sprite( 'dog-back-shadow.png' )
		self.interest = None
	class Move( object ):
		def __init__( self, object, dx, dy ):
			self.object = object
			self.phase = 0
			self.dx, self.dy = dx, dy
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			import math
			h = 40 * abs( jumpy( 2 * self.phase * math.pi / STEPTIME ) )
			s = 1 / (0.1 * h + 1)
			sprite.opacity = 255 * s
			glTranslatef( -h, -h/2, 0 )
			glTranslatef( self.dx * self.phase * GRIDX / fSTEPTIME, self.dy * self.phase * GRIDY / fSTEPTIME, 0 )
		def animate( self, sprite ):
			import math
			glTranslatef( self.dx * self.phase * GRIDX / fSTEPTIME, self.dy * self.phase * GRIDY / fSTEPTIME + 40 * abs( jumpy( 2 * self.phase * math.pi / STEPTIME ) ), 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase >= STEPTIME:
				self.object.visiblex += self.dx
				self.object.visibley += self.dy
				self.object.endanimation()
	class MoveLeft( Move ):
		def __init__( self, object ):
			Dog.Move.__init__( self, object, -1, 0 )
	class MoveRight( Move ):
		def __init__( self, object ):
			Dog.Move.__init__( self, object, 1, 0 )
	class MoveUp( Move ):
		def __init__( self, object ):
			Dog.Move.__init__( self, object, 0, 1 )
	class MoveDown( Move ):
		def __init__( self, object ):
			Dog.Move.__init__( self, object, 0, -1 )
	class Lick( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.starttime = object.world.time
			self.balloon = sprite( 'balloon-slurp.png' )
		def animate_shadow( self, sprite ):
			pass
		def animate( self, sprite ):
			if self.phase == 0:
				return
			import math
			glPushMatrix()
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			s = math.sin( self.phase * math.pi * 0.5 / fSTEPTIME ) ** 0.2
			glScalef( s, s, 1 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
			self.balloon.draw()
			glPopMatrix()
			glTranslatef( 0, 60 * abs( jumpy( self.phase * math.pi / STEPTIME ) ), 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase == 1:
				self.object.playsound( 'slurp.wav' )
			if self.phase >= STEPTIME * 2:
				self.object.endanimation()
	def hear( self, x, y, strength ):
		if self.interest is None or self.interest[2] < strength:
			self.interest = x, y, strength
	def step( self ):
		dx, dy = facing_to_offset( self.facing )
		o = self.world.map.get( (self.x+dx, self.y+dy), None )
		if o and isinstance( o, Child ) and not o.panicked:
			o.panicked = True
			o.justpanicked = True
			self.addanimation( 'Lick' )
			return

		if self.interest is not None:
			ix, iy, s = self.interest
			if abs( ix - self.x ) > abs( iy - self.y ):
				if ix < self.x:
					if self.facing == 'left':
						self.interest = None
					elif self.facing == 'right':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'left'
						self.addanimation( 'TurnLeft' )
						return
				else:
					if self.facing == 'right':
						self.interest = None
					elif self.facing == 'left':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'right'
						self.addanimation( 'TurnRight' )
						return
			else:
				if iy < self.y:
					if self.facing == 'down':
						self.interest = None
					elif self.facing == 'up':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'down'
						self.addanimation( 'TurnDown' )
						return
				else:
					if self.facing == 'up':
						self.interest = None
					elif self.facing == 'down':
						self.facing = facing_cw( self.facing )
						self.addanimation( 'Turn' + self.facing.capitalize() )
						return
					else:
						self.facing = 'up'
						self.addanimation( 'TurnUp' )
						return

		r = random.randint( 1, 100 )
		dx, dy = facing_to_offset( self.facing )
		if self.world.free_and_lawn( self.x + dx, self.y + dy ) and r <= 80:
			self.x += dx
			self.y += dy
			self.addanimation( 'Move' + self.facing.capitalize() )
			return
		dx, dy = facing_to_offset( facing_cw( self.facing ) )
		afree = self.world.free_and_lawn( self.x + dx, self.y + dy )
		achild = isinstance( self.world.map.get( (self.x + dx, self.y + dy), None ), Child )
		dx, dy = facing_to_offset( facing_ccw( self.facing ) )
		bfree = self.world.free_and_lawn( self.x + dx, self.y + dy )
		bchild = isinstance( self.world.map.get( (self.x + dx, self.y + dy), None ), Child )
		if achild != bchild:
			if achild:
				self.facing = facing_cw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
			else:
				self.facing = facing_ccw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
		elif afree != bfree:
			if afree:
				self.facing = facing_cw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
			else:
				self.facing = facing_ccw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
		else:
			r = random.randint( 1, 100 )
			if r <= 50:
				self.facing = facing_cw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )
			else:
				self.facing = facing_ccw( self.facing )
				self.addanimation( 'Turn' + self.facing.capitalize() )

class Ball( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.down = sprite( 'ball.png' )
		self.down_shadow = sprite( 'ball-shadow.png' )
	class Kicked( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.tx, self.ty = object.x, object.y
			self.starttime = object.world.time
		def getpos( self ):
			import math
			if self.phase < STEPTIME / 2:
				return 0, 0
			elif self.phase < 3 * STEPTIME / 2:
				p = (self.phase - STEPTIME / 2) / fSTEPTIME
				x = 0.7 * p
				y = 40 * jumpy( math.pi * p )
			elif self.phase < 3 * STEPTIME / 2 + 3 * STEPTIME / 10:
				p = (self.phase - STEPTIME / 2 - STEPTIME) / (3 * fSTEPTIME / 10)
				x = 0.7 + 0.2 * p
				y = 10 * jumpy( math.pi * p )
			else:
				p = (self.phase - STEPTIME / 2 - STEPTIME - 3 * fSTEPTIME / 10) / (2 * fSTEPTIME / 10)
				x = 0.7 + 0.2 + 0.1 * p
				y = 5 * jumpy( math.pi * p )
			return x, y
		def animate_shadow( self, sprite ):
			x, y = self.getpos()
			s = 1 / (0.1 * y + 1)
			sprite.opacity = s * 255
			glTranslatef( -y, -y/2, 0 )
			glTranslatef( self.dx * x * GRIDX, self.dy * x * GRIDY, 0 )
		def animate( self, sprite ):
			x, y = self.getpos()
			glTranslatef( self.dx * x * GRIDX, self.dy * x * GRIDY + y, 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.dx = self.tx - self.object.visiblex
			self.dy = self.ty - self.object.visibley
			self.phase += 1
			if self.phase == 12:
				playsound( 'ballkick.wav' )
			if self.phase >= STEPTIME*2:
				self.object.visiblex = self.tx
				self.object.visibley = self.ty
				self.object.endanimation()

class Rock( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.opaque = True
		self.down = sprite( 'rocks.png' )
		self.down_shadow = sprite( 'rocks-shadow.png' )

class Piggy( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.down = sprite( 'piggy.png' )
		self.down_shadow = sprite( 'piggy-shadow.png' )
	def usedBy( self, child ):
		playsound( 'piggy.wav' )
		self.world.objects.remove( self )
		child.properties.inventory.append( self )
		self.world.step()
	class Thrown( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.tx, self.ty = object.x, object.y
			self.starttime = object.world.time
		def getpos( self ):
			import math
			if self.phase < STEPTIME:
				p = self.phase / fSTEPTIME
				x = 0.6 * p
				y = 70 * jumpy( math.pi * (p + 0.5) / 1.5 )
			elif self.phase < STEPTIME + 3 * STEPTIME / 5:
				p = (self.phase - STEPTIME) / (3 * fSTEPTIME / 5)
				x = 0.6 + 0.3 * p
				y = 30 * jumpy( math.pi * p )
			else:
				p = (self.phase - STEPTIME - 3 * fSTEPTIME / 5) / (2 * fSTEPTIME / 5)
				x = 0.6 + 0.3 + 0.1 * p
				y = 10 * jumpy( math.pi * p )
			return x, y
		def animate_shadow( self, sprite ):
			x, y = self.getpos()
			s = 1 / (0.1 * y + 1)
			sprite.opacity = s * 255
			glTranslatef( -y, -y/2, 0 )
			glTranslatef( self.dx * x * GRIDX, self.dy * x * GRIDY, 0 )
		def animate( self, sprite ):
			x, y = self.getpos()
			glTranslatef( self.dx * x * GRIDX, self.dy * x * GRIDY + y, 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.dx = self.tx - self.object.visiblex
			self.dy = self.ty - self.object.visibley
			self.phase += 1
			if self.phase == 22:
				playsound( 'piggy.wav' )
			if self.phase >= 2*STEPTIME:
				self.object.visiblex = self.tx
				self.object.visibley = self.ty
				self.object.endanimation()

class Tap( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.up = sprite( 'tap-up.png' )
		self.down = sprite( 'tap-down.png' )
		self.left = sprite( 'tap-left.png' )
		self.right = sprite( 'tap-right.png' )
	def drawshadow( self ):
		pass
	def usedBy( self, child ):
		playsound( 'tap.wav' )
		if self.water in self.world.objects:
			self.world.objects.remove( self.water )
			self.world.noncolliders.append( self.water )
			for b in self.water.blocks:
				self.world.objects.remove( b )
			self.water.addanimation( 'TurnOff' )
		else:
			self.world.objects.append( self.water )
			self.world.noncolliders.remove( self.water )
			for b in self.water.blocks:
				self.world.objects.append( b )
				for c in self.world.children:
					if c.x == b.x and c.y == b.y:
						c.panicked = True
						c.justpanicked = True
			self.water.addanimation( 'TurnOn' )
		self.world.step()

class Water( Actor ):
	def __init__( self, world ):
		Actor.__init__( self, world )
		self.sprites = [sprite( 'water-1.png' ), sprite( 'water-1.png' ), sprite( 'water-2.png' ), sprite( 'water-2.png' ), sprite( 'water-3.png' ), sprite( 'water-3.png' )]
		self.down = self.sprites[0]
		self.timedivider = 1
		self.visible = True
	def draw( self ):
		if self.visible:
			glPushMatrix()
			glTranslatef( GRIDX, GRIDY, 0 )
			Actor.draw( self )
			glPopMatrix()
	def drawsprite( self ):
		self.sprites[0].draw()
		self.sprites[1].draw()
	def think( self ):
		Actor.think( self )
		self.timedivider -= 1
		if self.timedivider != 0: return
		self.timedivider = 2
		self.sprites = self.sprites[1:] + self.sprites[:1]
		self.down = self.sprites[0]
	def drawshadow( self ):
		pass
	class TurnOff( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			pass
		def animate( self, sprite ):
			s = 1 - self.phase / fSTEPTIME
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glScalef( s, s, 0 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.phase += 1
			if self.phase >= STEPTIME:
				self.object.endanimation()
				self.object.visible = False
	class TurnOn( object ):
		def __init__( self, object ):
			self.object = object
			self.phase = 0
			self.starttime = object.world.time
		def animate_shadow( self, sprite ):
			pass
		def animate( self, sprite ):
			s = self.phase / fSTEPTIME
			glTranslatef( sprite.width/2, sprite.height/2, 0 )
			glScalef( s, s, 0 )
			glTranslatef( -sprite.width/2, -sprite.height/2, 0 )
		def think( self ):
			if self.object.world.visibletime < self.starttime: return
			self.object.visible = True
			self.phase += 1
			if self.phase >= STEPTIME:
				self.object.endanimation()

class InvisibleBlock( Actor ):
	# invisible roadblock to be used as part of a larger sprite, like the water for a sprinkler
	def __init__( self, world ):
		Actor.__init__( self, world )
	def draw( self ):
		pass
	def drawshadow( self ):
		pass

class ChildProperties( object ):
	def __init__( self, name, **kwargs ):
		self.name = name
		self.abilities = set()
		self.sight = 6
		self.noise = 4
		self.passing = 3
		self.kind = 1
		self.inventory = []
		self.gender = 'boy'
		for k, v in kwargs.iteritems():
			setattr( self, k, v )
	def instanciate( self, world ):
		if self.gender == 'boy':
			return Boy( world, self )
		elif self.gender == 'girl':
			return Girl( world, self )

def facing_to_offset( f ):
	if f == 'up':
		return 0, 1
	elif f == 'down':
		return 0, -1
	elif f == 'left':
		return -1, 0
	elif f == 'right':
		return 1, 0
	assert False

def offset_to_facing( x, y ):
	if x == 0 and y == 1:
		return 'up'
	elif x == 0 and y == -1:
		return 'down'
	elif x == -1 and y == 0:
		return 'left'
	elif x == 1 and y == 0:
		return 'right'
	assert False

def perpendicular_facing( f ):
	if f == 'up':
		return 'left', 'right'
	elif f == 'down':
		return 'right', 'left'
	elif f == 'left':
		return 'up', 'down'
	elif f == 'right':
		return 'down', 'up'
	assert False

FACINGS = ['up', 'right', 'down', 'left']
def facing_cw( f ):
	return FACINGS[FACINGS.index( f ) - 3]
def facing_ccw( f ):
	return FACINGS[FACINGS.index( f ) - 1]

class Action( Phase ):
	def __init__( self, game ):
		self.game = game
		self.world = game.world
		self.deployment = game.world.deployment
		self.batch = pyglet.graphics.Batch()
		self.sprites = []
		self.objects = []
		self.noncolliders = []
		self.children = []
		self.createcurb()
		self.zoom = self.zoomtarget = 0.5
		self.offsetx, self.offsety = 0, 0
		self.mousex, self.mousey = 100, 100
		self.children.append( ChildProperties( 'Dennis', kind = 1 ).instanciate( self ) )
		self.children[-1].properties.abilities.add( 'sidestep' )
		self.children[-1].properties.abilities.add( 'backstep' )
		self.children[-1].setpos( 9, 1 )
		self.children[-1].setfacing( 'left' )
		self.children.append( ChildProperties( 'Jimmy', kind = 3 ).instanciate( self ) )
		self.children[-1].properties.abilities.add( 'sidestep' )
		self.children[-1].properties.abilities.add( 'backstep' )
		self.children[-1].setpos( 4, 1 )
		self.children[-1].setfacing( 'left' )
		self.children.append( ChildProperties( 'Johnny', kind = 2 ).instanciate( self ) )
		self.children[-1].properties.abilities.add( 'sidestep' )
		self.children[-1].properties.abilities.add( 'backstep' )
		self.children[-1].setpos( 1, 2 )
		self.children[-1].setfacing( 'right' )
		self.children.append( ChildProperties( 'Sue', kind = 1, gender = 'girl' ).instanciate( self ) )
		self.children[-1].properties.passing = 0
		self.children[-1].properties.abilities.add( 'sidestep' )
		self.children[-1].properties.abilities.add( 'backstep' )
		self.children[-1].setpos( 9, 4 )
		self.children[-1].setfacing( 'right' )
		self.selected = None
		self.select( self.children[-1] )
		self.selected.selected = True
		self.objects.extend( self.children )
		self.loadmap( '''
 RDR
    HHH
  SHHHH
    H  R
   RHE R
    H  R
   RH
 THHH
 BP
  P
''' )
		self.animating = 0
		self.time = self.visibletime = 0
		self.showvisions = []
		self.targetoffset = None
		self.dragged = False
		self.messages = []
		self.sighticon = sprite( 'eye-small.png' )
		self.passicon = sprite( 'ball-small.png' )
		self.noiseicon = sprite( 'noise-small.png' )
		self.portraitlabel = pyglet.text.Label( '', 'Tuffy', 18, bold=True, color=(255, 255, 255, 255), x=68, y=self.game.window.height-115, anchor_x='center', anchor_y='center' )
		self.tooltiplabel = pyglet.text.Label( '', 'Tuffy', 16, bold=True, color=(255, 255, 255, 255), anchor_x='left', anchor_y='center' )
		self.cursor = None
		self.bookkeeping()
		for o in self.objects:
			if not o.visible:
				o.opacity = 0
	def init( self ):
		glClearColor( 92/255.0, 92/255.0, 92/255.0, 1 )
	def addmessage( self, message ):
		if len( self.messages ) == 0:
			y = self.game.window.height/2
		else:
			y = self.messages[-1].y - 40
		self.messages.append( pyglet.text.Label( message, 'Tuffy', 18, bold=True, color=(255, 255, 255, 255), x=self.game.window.width/2, y=y, anchor_x='center', anchor_y='center' ) )
	def loadmap( self, mapstr ):
		lines = mapstr.split( '\n' )[1:-1]
		map = {}
		for y1 in range( len( lines ) ):
			line = lines[y1]
			y = self.deployment.height - y1
			for x1 in range( len( line ) ):
				c = line[x1]
				x = x1 + 1
				if c == ' ':
					continue
				map[x, y] = c
		def add( cls, x, y ):
			obj = cls( self )
			obj.setpos( x, y )
			self.objects.append( obj )
			return obj
		def addsprite( f, x, y ):
			s = sprite( f, batch=self.batch, mipmap=False )
			s.x, s.y = x * GRIDX, y * GRIDY
			self.sprites.append( s )
			return s

		hoses = {}
		def joinhose( *ps ):
			union = set()
			for p in ps:
				if p in hoses:
					union |= hoses[p]
				else:
					union.add( p )
			tap = None
			sprinkler = None
			for p in union:
				hoses[p] = union
				if map[p] == 'S':
					sprinkler = p
				elif map[p] == 'T':
					tap = p
			if tap and sprinkler:
				for a in self.objects:
					if isinstance( a, Water ) and (a.x+1, a.y+1) == sprinkler:
						sprinkler = a
					elif isinstance( a, Tap ) and (a.x, a.y) == tap:
						tap = a
				if isinstance( tap, Tap ) and isinstance( sprinkler, Water ):
					tap.water = sprinkler
		for x, y in map:
			c = map[x, y]
			if c == 'B':
				add( Ball, x, y )
			elif c == 'E':
				add( Elderly, x, y )
			elif c == 'D':
				add( Dog, x, y )
			elif c == 'R':
				add( Rock, x, y )
			elif c == 'P':
				add( Piggy, x, y )
			elif c in 'HST':
				thisp  = x, y
				leftp  = x-1, y
				rightp = x+1, y
				upp    = x, y+1
				downp  = x, y-1
				left  = map.get( leftp , 'X' )
				right = map.get( rightp, 'X' )
				up    = map.get( upp   , 'X' )
				down  = map.get( downp , 'X' )
				if c == 'H':
					if left in 'HST' and right in 'HST' and up in 'HST' and down in 'HST':
						addsprite( 'hose-cross.png', x, y )
						joinhose( leftp, rightp )
						joinhose( upp, downp )
					elif left in 'HST' and right in 'HST':
						addsprite( 'hose-left-right.png', x, y )
						joinhose( leftp, thisp, rightp )
					elif up in 'HST' and down in 'HST':
						addsprite( 'hose-up-down.png', x, y )
						joinhose( upp, thisp, downp )
					elif left in 'HST' and down in 'HST':
						addsprite( 'hose-left-down.png', x, y )
						joinhose( leftp, thisp, downp )
					elif left in 'HST' and up in 'HST':
						addsprite( 'hose-left-up.png', x, y )
						joinhose( leftp, thisp, upp )
					elif right in 'HST' and down in 'HST':
						addsprite( 'hose-right-down.png', x, y )
						joinhose( rightp, thisp, downp )
					elif right in 'HST' and up in 'HST':
						addsprite( 'hose-right-up.png', x, y )
						joinhose( rightp, thisp, upp )
				elif c == 'S':
					if right in 'HST':
						addsprite( 'sprinkler-left.png', x, y )
						joinhose( rightp, thisp )
					elif left in 'HST':
						addsprite( 'sprinkler-right.png', x, y )
						joinhose( leftp, thisp )
					elif up in 'HST':
						addsprite( 'sprinkler-down.png', x, y )
						joinhose( upp, thisp )
					elif down in 'HST':
						addsprite( 'sprinkler-up.png', x, y )
						joinhose( downp, thisp )
					water = add( Water, x-1, y-1 )
					water.blocks = []
					for i in range( 3 ):
						for j in range( 3 ):
							block = add( InvisibleBlock, x-1+i, y-1+j )
							water.blocks.append( block )
				elif c == 'T':
					tap = add( Tap, x, y )
					if right in 'HST':
						tap.setfacing( 'left' )
						joinhose( rightp, thisp )
					elif left in 'HST':
						tap.setfacing( 'right' )
						joinhose( leftp, thisp )
					elif up in 'HST':
						tap.setfacing( 'down' )
						joinhose( upp, thisp )
					elif down in 'HST':
						tap.setfacing( 'up' )
						joinhose( downp, thisp )
			else:
				assert False, 'unknown map symbol: ' + c
	def createcurb( self ):
		# green patch
		w = (self.deployment.width+2) * GRIDX
		h = (self.deployment.height+2) * GRIDY
		self.batch.add( 4, GL_QUADS, None, ('v2i', (0,0, w,0, w,h, 0,h)), ('c3B', (25,123,48, 25,123,48, 25,123,48, 25,123,48)) )

		# TODO: Mipmapped curb tiles result in subpixel gaps. Disabled mipmaps for now.
		s = sprite( 'curb-topleft.png', batch=self.batch, mipmap=False )
		s.x, s.y = 0, (self.deployment.height+1) * GRIDY
		self.sprites.append( s )
		for i in range( self.deployment.width ):
			s = sprite( 'curb-top.png', batch=self.batch, mipmap=False )
			s.x, s.y = (i+1) * GRIDX, (self.deployment.height+1) * GRIDY
			self.sprites.append( s )
		s = sprite( 'curb-topright.png', batch=self.batch, mipmap=False )
		s.x, s.y = (self.deployment.width+1) * GRIDX, (self.deployment.height+1) * GRIDY
		self.sprites.append( s )
		for i in range( self.deployment.width ):
			s = sprite( 'curb-right.png', batch=self.batch, mipmap=False )
			s.x, s.y = (self.deployment.width+1) * GRIDX, (i+1) * GRIDY
			self.sprites.append( s )
		s = sprite( 'curb-bottomright.png', batch=self.batch, mipmap=False )
		s.x, s.y = (self.deployment.width+1) * GRIDX, 0
		self.sprites.append( s )
		for i in range( self.deployment.width ):
			s = sprite( 'curb-bottom.png', batch=self.batch, mipmap=False )
			s.x, s.y = (i+1) * GRIDX, 0
			self.sprites.append( s )
		s = sprite( 'curb-bottomleft.png', batch=self.batch, mipmap=False )
		s.x, s.y = 0, 0
		self.sprites.append( s )
		for i in range( self.deployment.width ):
			s = sprite( 'curb-left.png', batch=self.batch, mipmap=False )
			s.x, s.y = 0, (i+1) * GRIDY
			self.sprites.append( s )

	def draw( self ):
		glPushMatrix()
		glTranslatef( self.offsetx, self.offsety, 0 )
		glScalef( self.zoom, self.zoom, 1 )
		self.batch.draw()
		for vision in self.showvisions:
			vision.draw()
		for o in self.objects:
			o.drawshadow()
		for o in self.noncolliders:
			o.drawshadow()
		for o in self.objects:
			o.draw()
		for o in self.noncolliders:
			o.draw()
		if self.cursor is not None:
			self.cursor.draw()
		glPopMatrix()
		if self.selected is not None:		# HUD
			s = self.selected
			glPushMatrix()
			glTranslatef( -60, self.game.window.height - 190, 0 )
			s.portrait.draw()
			glPopMatrix()
			glPushMatrix()
			glTranslatef( 10, self.game.window.height - 180, 0 )
			for i in range( s.properties.sight ):
				self.sighticon.draw()
				glTranslatef( 0, -20, 0 )
			glPopMatrix()
			glPushMatrix()
			glTranslatef( 40, self.game.window.height - 180, 0 )
			for i in range( s.properties.passing ):
				self.passicon.draw()
				glTranslatef( 0, -20, 0 )
			glPopMatrix()
			glPushMatrix()
			glTranslatef( 70, self.game.window.height - 180, 0 )
			for i in range( s.properties.noise ):
				self.noiseicon.draw()
				glTranslatef( 0, -20, 0 )
			glPopMatrix()
			glPushMatrix()
			glTranslatef( -200, self.game.window.height - 650, 0 )
			for i in s.properties.inventory:
				i.drawsprite()
				glTranslatef( 0, -30, 0 )
			glPopMatrix()
			if self.portraitlabel.text != s.properties.name:
				self.portraitlabel.text = s.properties.name
			self.portraitlabel.draw()
		for l in self.messages:
			l.draw()
		if self.game.option_tooltips:
			self.tooltiplabel.draw()
	def think( self ):
		if self.zoom != self.zoomtarget:
			oldzoom = self.zoom
			self.zoom *= (self.zoomtarget / self.zoom) ** 0.2
			if self.zoom > self.zoomtarget > oldzoom or self.zoom < self.zoomtarget < oldzoom:
				self.zoom = self.zoomtarget
			self.offsetx = self.mousex - (self.mousex - self.offsetx) * self.zoom / oldzoom
			self.offsety = self.mousey - (self.mousey - self.offsety) * self.zoom / oldzoom
		if self.targetoffset is not None:
			tx, ty = self.targetoffset
			self.offsetx += 0.1 * (tx - self.offsetx)
			self.offsety += 0.1 * (ty - self.offsety)
		if self.animating > 0:
			self.animating -= 1
			if self.animating % STEPTIME == 0:
				self.visibletime += 1
		elif self.selected is None:
			self.step()
		if self.game.option_edgescroll:
			SCROLLBAND = 30
			if self.mousex < SCROLLBAND:
				self.offsetx += SCROLLBAND - self.mousex
				self.targetoffset = None
			if self.mousex > self.game.window.width - SCROLLBAND:
				self.offsetx -= SCROLLBAND - self.game.window.width + self.mousex
				self.targetoffset = None
			if self.mousey < SCROLLBAND:
				self.offsety += SCROLLBAND - self.mousey
				self.targetoffset = None
			if self.mousey > self.game.window.height - SCROLLBAND:
				self.offsety -= SCROLLBAND - self.game.window.height + self.mousey
				self.targetoffset = None
		for o in self.objects:
			o.think()
		for o in self.noncolliders:
			o.think()
		for vision in self.showvisions:
			vision.think()
		for l in self.messages[:]:
			alpha = l.color[-1]
			if alpha > 200:
				l.color = 255, 255, 255, alpha-1
			else:
				l.color = 255, 255, 255, alpha-5
			if alpha <= 5:
				self.messages.remove( l )
	def noise( self, x, y, range ):
		for o in self.objects:
			if hasattr( o, 'hear' ) and max( abs( o.x - x ), abs( o.y - y ) ) <= range:
				o.hear( x, y, range - max( abs( o.x - x ), abs( o.y - y ) ) )
	def on_mouse_scroll( self, x, y, scroll_x, scroll_y ):
		for i in range( scroll_y ):
			self.zoomtarget /= 0.9
		for i in range( -scroll_y ):
			self.zoomtarget *= 0.9
		if self.zoomtarget > 1:
			self.zoomtarget = 1
		elif self.zoomtarget < 0.2:
			self.zoomtarget = 0.2
		self.targetoffset = None
		self.tooltiplabel.text = ''
	def on_mouse_drag( self, x, y, dx, dy, button, modifiers ):
		self.dragged = True
		self.offsetx += dx
		self.offsety += dy
		self.targetoffset = None
	def select( self, next ):
		if self.selected is not None:
			self.selected.selected = False
		self.selected = next
		if self.selected is not None:
			self.selected.selected = True
	def switchselection( self, dir = 1 ):
		usable = [c for c in self.children if c.can_act()]
		if self.selected in usable:
			self.select( usable[(usable.index( self.selected ) + dir)%len( usable )] )
		elif len( usable ) > 0:
			self.select( usable[0] )
		else:
			self.select( None )
		if self.selected is not None:
			tx = self.game.window.width / 2 - (self.selected.x + 0.5) * GRIDX * self.zoom
			ty = self.game.window.height / 2 - (self.selected.y + 0.5) * GRIDY * self.zoom
			dx = tx - self.offsetx
			dy = ty - self.offsety
			cx = 0.45 * self.game.window.width
			cy = 0.3 * self.game.window.height
			if abs( dx ) > cx:
				dx *= (abs( dx ) - cx) / abs( dx )
			else:
				dx = 0
			if abs( dy ) > cy:
				dy *= (abs( dy ) - cy) / abs( dy )
			else:
				dy = 0
			if dx != 0 or dy != 0:
				self.targetoffset = self.offsetx + dx, self.offsety + dy
	def mouse_to_offset( self, x, y ):
		return (x - self.offsetx) / self.zoom, (y - self.offsety) / self.zoom
	def mouse_to_game( self, x, y ):
		ox, oy = self.mouse_to_offset( x, y )
		# fix for negative values, which do not really occur anyway...
		if ox < 0: ox -= GRIDX
		if oy < 0: oy -= GRIDY
		return int( ox/GRIDX ), int( oy/GRIDY )
	def game_to_mouse( self, x, y ):
		return x * GRIDX * self.zoom + self.offsetx, y * GRIDY * self.zoom + self.offsety
	def on_mouse_motion( self, x, y, dx, dy ):
		if self.cursor is not None:
			self.cursor.moved( x, y )
		self.mousex, self.mousey = x, y
		gx, gy = self.mouse_to_game( x, y )
		for o in self.objects:
			if o.x == gx and o.y == gy and hasattr( o, 'sight' ) and o.visible:
				if self.showvisions and o is self.showvisions[-1].actor:
					break
				else:
					if self.showvisions:
						self.showvisions[-1].close()
					self.showvisions.append( Action.ShowVision( o ) )
					break
		else:
			if self.showvisions:
				self.showvisions[-1].close()

		tip = self.tooltiplabel
		if self.selected and 25 < x < 80 and 260 > y > 230 - 30 * len( self.selected.properties.inventory ):
			inventory = self.selected.properties.inventory
			i = (245 - y) / 30
			if i < 0: i = 0
			if i >= len( inventory ): i = len( inventory ) - 1
			if i > 0 and isinstance( inventory[i], Piggy ):
				tip.text = 'Click to throw piggy'
				tip.x, tip.y = 70, 235 - i * 30
				return
		if self.selected and 25 < x < 50 and 500 > y > 260:
			tip.text = 'Sight radius (%d)'%self.selected.properties.sight
			tip.x, tip.y = 130, 460
			return
		if self.selected and 60 < x < 85 and 500 > y > 260:
			tip.text = 'Passing radius (%d)'%self.selected.properties.passing
			tip.x, tip.y = 130, 460
			return
		if self.selected and 95 < x < 120 and 500 > y > 260:
			tip.text = 'Noise radius (%d)'%self.selected.properties.noise
			tip.x, tip.y = 130, 460
			return
		gx, gy = self.mouse_to_game( x, y )
		o = self.map.get( (gx, gy), None )
		tip.x, tip.y = self.game_to_mouse( gx, gy )
		tip.y += 100 * self.zoom
		tip.x += 180 * self.zoom
		if o is None or not o.visible:
			tip.text = ''
		elif isinstance( o, Child ):
			tip.text = o.properties.name
			if o.panicked:
				tip.text += ', panicked'
			elif getattr( o, 'talkingto', False ):
				tip.text += ', engaged in smalltalk'
		elif  isinstance( o, Piggy ):
			tip.text = 'Rubber piggy, a useful distraction'
		elif  isinstance( o, Ball ):
			tip.text = 'Soccer ball, the one you came to rescue'
		elif  isinstance( o, Dog ):
			tip.text = 'Don\'t let him lick your face!'
		elif  isinstance( o, Tap ):
			tip.text = 'Walk here to turn tap'
		elif  isinstance( o, InvisibleBlock ):
			tip.text = 'Can not pass through water'
		else:
			tip.text = ''

	class ShowVision( object ):
		def __init__( self, actor ):
			self.actor = actor
			self.batches = [pyglet.graphics.Batch() for i in range( 3 )]
			self.sprites = []
			for x, y in actor.vision():
				for k in range( 3 ):
					s = sprite( 'eye-%d.png'%(k+1), batch = self.batches[k] )
					s.x = GRIDX * x - s.width / 2 + GRIDX / 2
					s.y = GRIDY * y - s.height / 2 + GRIDY / 2
					self.sprites.append( s )
			self.state = 0
			self.closing = False
		def close( self ):
			self.closing = True
		def think( self ):
			if not self.closing:
				if self.state < 4:
					self.state += 1
			else:
				self.state -= 1
				if self.state < 0:
					self.actor.world.showvisions.remove( self )
		def draw( self ):
			self.batches[self.state/2].draw()
	class BallPass( object ):
		def __init__( self, world, ball ):
			self.world = world
			self.ball = ball
			self.sprite = sprite( 'ball.png' )
			self.sprite.opacity = 0
		def draw( self ):
			self.sprite.draw()
		def clicked( self, x, y ):
			gx, gy = self.world.mouse_to_game( x, y )
			s = self.world.selected
			if gx == self.ball.x and gy == self.ball.y:
				self.world.cursor = None
				self.world.addmessage( 'Pass cancelled.' )
			elif self.acceptable( gx, gy ):
				self.world.cursor = None
				s.addanimation( 'Use' + s.facing.capitalize() )
				self.ball.x = gx
				self.ball.y = gy
				self.ball.addanimation( 'Kicked' )
				self.world.noise( gx, gy, 3 )
				self.world.step()
			elif self.possible( gx, gy ):
				self.world.addmessage( 'Must pass to an able recipient.' )
				self.world.addmessage( '(Or click ball to cancel.)' )
			else:
				self.world.addmessage( 'Target must be in line of sight and passing range (%d for %s).'%(s.properties.passing, s.properties.name) )
		def moved( self, x, y ):
			gx, gy = self.world.mouse_to_game( x, y )
			if self.possible( gx, gy ):
				self.sprite.opacity = 100
			else:
				self.sprite.opacity = 0
			self.sprite.x = gx * GRIDX - self.sprite.width / 2 + GRIDX / 2
			self.sprite.y = gy * GRIDY - self.sprite.height / 2 + GRIDY / 2
		def acceptable( self, x, y ):
			if not self.possible( x, y ):
				return False
			for c in self.world.children:
				dx, dy = facing_to_offset( c.facing )
				if c.x + dx == x and c.y + dy == y and not c.panicked:
					return True
			else:
				return False
		def possible( self, x, y ):
			if x == self.ball.x and y == self.ball.y:
				return False
			if not self.world.free( x, y ):
				return False
			s = self.world.selected
			if (x, y) not in list( s.vision() ):
				return False
			if max( abs( x - s.x ), abs( y - s.y ) ) > s.properties.passing:
				return False
			return True
	class ThrowPiggy( object ):
		def __init__( self, world, piggy ):
			self.world = world
			self.piggy = piggy
			self.sprite = sprite( 'piggy.png' )
			self.sprite.opacity = 0
		def draw( self ):
			self.sprite.draw()
		def clicked( self, x, y ):
			gx, gy = self.world.mouse_to_game( x, y )
			s = self.world.selected
			if self.acceptable( gx, gy ):
				self.world.cursor = None
				s.properties.inventory.remove( self.piggy )
				self.piggy.visiblex = s.x
				self.piggy.visibley = s.y
				self.world.objects.append( self.piggy )
				self.piggy.x = gx
				self.piggy.y = gy
				self.piggy.addanimation( 'Thrown' )
				self.world.noise( gx, gy, 10 )
				self.world.step()
			else:
				self.world.addmessage( 'Target must be in line of sight and passing range (%d for %s).'%(s.properties.passing, s.properties.name) )
				self.world.addmessage( 'Throw cancelled.' )
				self.world.cursor = None
		def moved( self, x, y ):
			gx, gy = self.world.mouse_to_game( x, y )
			if self.acceptable( gx, gy ):
				self.sprite.opacity = 100
			else:
				self.sprite.opacity = 0
			self.sprite.x = gx * GRIDX - self.sprite.width / 2 + GRIDX / 2
			self.sprite.y = gy * GRIDY - self.sprite.height / 2 + GRIDY / 2
		def acceptable( self, x, y ):
			if not self.world.free( x, y ):
				return False
			s = self.world.selected
			if (x, y) not in list( s.vision() ):
				return False
			if max( abs( x - s.x ), abs( y - s.y ) ) > s.properties.passing:
				return False
			return True
	def on_mouse_release( self, x, y, button, modifiers ):
		if self.dragged:
			self.dragged = False
			return
		if self.animating:
			return
		if self.cursor is not None:
			self.cursor.clicked( x, y)
			return
		if self.selected and 25 < x < 80 and 260 > y > 230 - 30 * len( self.selected.properties.inventory ):
			inventory = self.selected.properties.inventory
			i = (245 - y) / 30
			if i < 0: i = 0
			if i >= len( inventory ): i = len( inventory ) - 1
			if i > 0 and isinstance( inventory[i], Piggy ):
				self.cursor = Action.ThrowPiggy( self, inventory[i] )
				self.addmessage( 'Select throw target!' )
				return
		gx, gy = self.mouse_to_game( x, y )
		o = self.map.get( (gx, gy), None )
		dx, dy = facing_to_offset( self.selected.facing )
		if isinstance( o, Ball ) and self.selected.x + dx == gx and self.selected.y + dy == gy:
			if self.selected.properties.passing > 0:
				self.cursor = Action.BallPass( self, o )
				self.addmessage( 'Select pass target!' )
			else:
				self.addmessage( '%s has no passing ability!'%self.selected.properties.name )
			return
		mind = None
		for c in self.children:
			if not c.can_act(): continue
			d = (c.x - gx) ** 2 + (c.y - gy) ** 2
			if mind is None or d < mind:
				mind = d
				b = c
		if mind is not None and b is not self.selected:
			self.select( b )
			playsound( 'switch.wav' )
	def on_key_press( self, symbol, modifiers ):
		if self.animating:
			return
		self.cursor = None
		if symbol == pyglet.window.key.TAB:
			if not modifiers:
				self.switchselection()
				playsound( 'switch.wav' )
			elif modifiers & pyglet.window.key.MOD_SHIFT:
				self.switchselection( -1 )
				playsound( 'switch.wav' )
		s = self.selected
		if s is not None and s.can_act():
			if not modifiers:
				if symbol == pyglet.window.key.RIGHT:
					if s.facing == 'left':
						s.facing = 'down'
						s.addanimation( 'TurnDown' )
						self.step()
					elif s.facing != 'right':
						s.facing = 'right'
						s.addanimation( 'TurnRight' )
						self.step()
					elif self.free( s.x+1, s.y ):
						s.x += 1
						s.addanimation( 'MoveRight' )
						self.noise( s.x, s.y, s.properties.noise )
						self.step()
					elif self.usable( s.x+1, s.y ):
						s.addanimation( 'UseRight' )
						o = self.map[s.x+1, s.y]
						o.usedBy( s )
				elif symbol == pyglet.window.key.LEFT:
					if s.facing == 'right':
						s.facing = 'up'
						s.addanimation( 'TurnUp' )
						self.step()
					elif s.facing != 'left':
						s.facing = 'left'
						s.addanimation( 'TurnLeft' )
						self.step()
					elif self.free( s.x-1, s.y ):
						s.x -= 1
						s.addanimation( 'MoveLeft' )
						self.noise( s.x, s.y, s.properties.noise )
						self.step()
					elif self.usable( s.x-1, s.y ):
						s.addanimation( 'UseLeft' )
						o = self.map[s.x-1, s.y]
						o.usedBy( s )
				elif symbol == pyglet.window.key.UP:
					if s.facing == 'down':
						s.facing = 'right'
						s.addanimation( 'TurnRight' )
						self.step()
					elif s.facing != 'up':
						s.facing = 'up'
						s.addanimation( 'TurnUp' )
						self.step()
					elif self.free( s.x, s.y+1 ):
						s.y += 1
						s.addanimation( 'MoveUp' )
						self.noise( s.x, s.y, s.properties.noise )
						self.step()
					elif self.usable( s.x, s.y+1 ):
						s.addanimation( 'UseUp' )
						o = self.map[s.x, s.y+1]
						o.usedBy( s )
				elif symbol == pyglet.window.key.DOWN:
					if s.facing == 'up':
						s.facing = 'left'
						s.addanimation( 'TurnLeft' )
						self.step()
					elif s.facing != 'down':
						s.facing = 'down'
						s.addanimation( 'TurnDown' )
						self.step()
					elif self.free( s.x, s.y-1 ):
						s.y -= 1
						s.addanimation( 'MoveDown' )
						self.noise( s.x, s.y, s.properties.noise )
						self.step()
					elif self.usable( s.x, s.y-1 ):
						s.addanimation( 'UseDown' )
						o = self.map[s.x, s.y-1]
						o.usedBy( s )
				elif symbol == pyglet.window.key.PERIOD:
					self.step()
			elif modifiers & pyglet.window.key.MOD_SHIFT:
				if symbol == pyglet.window.key.RIGHT and self.free( s.x+1, s.y ) and (s.facing == 'right' or s.facing == 'left' and 'backstep' in s.properties.abilities or s.facing in ['up', 'down'] and 'sidestep' in s.properties.abilities):
					s.x += 1
					s.addanimation( 'MoveRight' )
					self.noise( s.x, s.y, s.properties.noise )
					self.step()
				elif symbol == pyglet.window.key.LEFT and self.free( s.x-1, s.y ) and (s.facing == 'left' or s.facing == 'right' and 'backstep' in s.properties.abilities or s.facing in ['up', 'down'] and 'sidestep' in s.properties.abilities):
					s.x -= 1
					s.addanimation( 'MoveLeft' )
					self.noise( s.x, s.y, s.properties.noise )
					self.step()
				elif symbol == pyglet.window.key.UP and self.free( s.x, s.y+1 ) and (s.facing == 'up' or s.facing == 'down' and 'backstep' in s.properties.abilities or s.facing in ['left', 'right'] and 'sidestep' in s.properties.abilities):
					s.y += 1
					s.addanimation( 'MoveUp' )
					self.noise( s.x, s.y, s.properties.noise )
					self.step()
				elif symbol == pyglet.window.key.DOWN and self.free( s.x, s.y-1 ) and (s.facing == 'down' or s.facing == 'up' and 'backstep' in s.properties.abilities or s.facing in ['left', 'right'] and 'sidestep' in s.properties.abilities):
					s.y -= 1
					s.addanimation( 'MoveDown' )
					self.noise( s.x, s.y, s.properties.noise )
					self.step()
	def step( self ):
		for o in self.objects:
			self.remap()
			o.step()
		for vision in self.showvisions:
			vision.close()
		self.cursor = None
		self.animating += STEPTIME
		self.time += 1
		self.bookkeeping()
	def bookkeeping( self ):
		self.objects.sort( key = lambda o: (-o.y, -o.x) )
		self.remap()
		if self.game.option_childvision:
			for o in self.objects:
				if isinstance( o, Elderly ) or isinstance( o, Dog ):
					o.visible = False
			for c in self.children:
				for p in c.vision():
					o = self.map.get( p, None )
					if isinstance( o, Elderly ) or isinstance( o, Dog ):
						o.visible = True
	def remap( self ):
		self.map = {}
		for o in self.objects:
			self.map[o.x, o.y] = o
	def free( self, x, y ):
		return (x, y) not in self.map
	def usable( self, x, y ):
		return hasattr( self.map.get( (x, y), None ), 'usedBy' )
	def is_lawn( self, x, y ):
		return 0 < x <= self.deployment.width and 0 < y <= self.deployment.height
	def free_and_lawn( self, x, y ):
		return self.free( x, y ) and self.is_lawn( x, y )

class Menu( Phase ):
	showtitle = True
	def __init__( self, game ):
		self.game = game
		self.batch = pyglet.graphics.Batch()
		self.sprite = sprite( 'title.png', batch=self.batch, mipmap=False )
		if Menu.showtitle:
			self.titledelay = 40
		else:
			self.titledelay = 0
		self.menu = ['NEW GAME', 'CONTINUE', 'OPTIONS', 'CREDITS', 'QUIT GAME']
		self.labels = [pyglet.text.Label( self.menu[i], font_name='Tuffy', bold=True, font_size=20, color=(0, 0, 0, 255), x=120+i*190, y=30-self.titledelay*4, anchor_x='center', anchor_y='center', dpi=96 ) for i in range( len( self.menu ) )]
	def init( self ):
		glClearColor( 25/255.0, 123/255.0, 48/255.0, 1 )
		self.game.music.play()
	def end( self ):
		self.game.music.pause()
	def on_mouse_motion( self, x, y, dx, dy ):
		for l in self.labels:
			l.color = (0, 0, 0, 255)
		for l in self.labels:
			if l.x - l.content_width/2 <= x <= l.x + l.content_width/2 and l.y - l.content_height/2 <= y <= l.y + l.content_height/2:
				l.color = (255, 255, 255, 255)
				break
	def on_mouse_release( self, x, y, button, modifiers ):
		for l in self.labels:
			if l.x - l.content_width/2 <= x <= l.x + l.content_width/2 and l.y - l.content_height/2 <= y <= l.y + l.content_height/2:
				self.handle( l.text )
				break
	def handle( self, menu ):
		if menu == 'NEW GAME':
			self.game.world = World.random()
			self.game.setphase( Transition( self.game, self, Base( self.game ) ) )
		elif menu == 'CONTINUE':
			self.game.world = World.load()
			if self.game.world.deployment:
				self.game.setphase( Transition( self.game, self, Action( self.game ) ) )
			else:
				self.game.setphase( Transition( self.game, self, Base( self.game ) ) )
		elif menu == 'OPTIONS':
			self.game.setphase( Options( self.game ) )
		elif menu == 'CREDITS':
			self.game.setphase( Credits( self.game ) )
		elif menu == 'QUIT GAME':
			pyglet.app.exit()
	def draw( self ):
		self.batch.draw()
		for l in self.labels:
			l.draw()
	def think( self ):
		if self.titledelay > 0:
			self.titledelay -= 1
			for l in self.labels:
				l.y += 4
			if self.titledelay <= 0:
				Menu.showtitle = False

class Credits( Phase ):
	def __init__( self, game ):
		self.game = game
		self.batch = pyglet.graphics.Batch()
		self.sprite = sprite( 'untitle.png', batch=self.batch, mipmap=False )
		self.lines = ['Voice of Sue:', 'Beata Darabos', '', 'This font (called Tuffy):', 'Thatcher Ulrich', '', 'Everything else:', 'Daniel Darabos', 'darabos.daniel@gmail.com', '', 'Created for PyWeek #8', 'for the theme "Get Off My Lawn!"', '', 'This game is written in Python,', 'uses the Pyglet library', 'and is released under the GPL']
		self.labels = [pyglet.text.Label( self.lines[i], font_name='Tuffy', bold=True, font_size=20, x=self.game.window.width / 2, y=-40 * i, anchor_x='center', anchor_y='center', dpi=96 ) for i in range( len( self.lines ) )]
	def init( self ):
		glClearColor( 25/255.0, 123/255.0, 48/255.0, 1 )
		self.game.setmusic( 'bonus music.ogg' )
		self.game.music.eos_action = self.game.music.EOS_NEXT
		self.game.music.play()
	def end( self ):
		self.game.setmusic( 'guitar.ogg' )
	def on_mouse_release( self, x, y, button, modifiers ):
		self.game.setphase( Transition( self.game, self, Menu( self.game ) ) )
	def on_key_press( self, symbol, modifiers ):
		self.game.setphase( Transition( self.game, self, Menu( self.game ) ) )
	def draw( self ):
		self.batch.draw()
		for l in self.labels:
			l.draw()
	def think( self ):
		for l in self.labels:
			l.y += 1

class Options( Phase ):
	def __init__( self, game ):
		self.game = game
		self.batch = pyglet.graphics.Batch()
		self.sprite = sprite( 'untitle.png', batch=self.batch, mipmap=False )
		self.explanation = pyglet.text.Label( 'These are the current settings. Click to change.', font_name='Tuffy', bold=True, font_size=14, x = 80, y = 450 )
		self.settings = [('childvision', 'Dogs and the elderly only visible when looked at. (Hard)', 'Dogs and the elderly always visible. (Easy)'),
			('tooltips', 'Tooltips enabled', 'Tooltips disabled'),
			('edgescroll', 'Mouse moved to the edge of the window activates scrolling', 'Scrolling only via dragging'),
			]
		options = []
		for id, true, false in self.settings:
			if getattr( self.game, 'option_' + id ):
				options.append( true )
			else:
				options.append( false )
		options += ['', 'OK']
		self.labels = [pyglet.text.Label( options[i], font_name='Tuffy', bold=True, font_size=20, x=self.game.window.width / 2, y=350 - 50*i, anchor_x='center', anchor_y='center', dpi=96 ) for i in range( len( options ) )]
	def init( self ):
		glClearColor( 25/255.0, 123/255.0, 48/255.0, 1 )
		self.game.music.play()
	def end( self ):
		self.game.music.pause()
	def on_mouse_release( self, x, y, button, modifiers ):
		for l in self.labels:
			if l.x - l.content_width/2 <= x <= l.x + l.content_width/2 and l.y - l.content_height/2 <= y <= l.y + l.content_height/2:
				self.handle( l )
				break
	def handle( self, menu ):
		for id, true, false in self.settings:
			if menu.text == true:
				setattr( self.game, 'option_' + id, False )
				menu.text = false
			elif menu.text == false:
				setattr( self.game, 'option_' + id, True )
				menu.text = true
		if menu.text == 'OK':
			# TODO
			#self.game.saveoptions()
			self.game.setphase( Transition( self.game, self, Menu( self.game ) ) )
	def draw( self ):
		self.batch.draw()
		self.explanation.draw()
		for l in self.labels:
			l.draw()

def main():
	game = Game()
#	game.world = World.random()
#	game.setphase( Base( game ) )
	game.start()
